"""
Transformers (aka data mappers) are used to transform objects from one class to another
using a transformation specification.
"""

import logging
from abc import ABC
from copy import deepcopy
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, Optional, Union

import yaml
from curies import Converter
from linkml_runtime import SchemaView
from linkml_runtime.processing.referencevalidator import ReferenceValidator
from linkml_runtime.utils.introspection import package_schemaview
from linkml_runtime.utils.yamlutils import YAMLRoot
from pydantic import BaseModel

from linkml_transformer.datamodel.transformer_model import (
    ClassDerivation,
    CollectionType,
    EnumDerivation,
    SlotDerivation,
    TransformationSpecification,
)
from linkml_transformer.inference.inference import induce_missing_values

logger = logging.getLogger(__name__)


OBJECT_TYPE = Union[Dict[str, Any], BaseModel, YAMLRoot]
"""An object can be a plain python dict, a pydantic object, or a linkml YAMLRoot"""


@dataclass
class Transformer(ABC):
    """
    Base class for all transformers.

    A transformer will generate an instance of a target class from
    an instance of a source class, making use of a specification.

    This is an abstract class. Different implementations will
    subclass this
    """

    specification: TransformationSpecification = None
    """A specification of how to generate target objects from source objects."""

    source_schemaview: SchemaView = None
    """A view over the schema describing the input/source object."""

    _derived_specification: TransformationSpecification = None
    """A specification with inferred missing values."""

    target_schemaview: Optional[SchemaView] = None
    """A view over the schema describing the output/target object."""

    unrestricted_eval: bool = field(default=False)
    """Set to True to allow arbitrary evals as part of transformation."""

    _curie_converter: Converter = None

    def map_object(self, obj: OBJECT_TYPE, source_type: str = None, **kwargs) -> OBJECT_TYPE:
        """
        Transform source object into an instance of the target class.

        :param obj:
        :param source_type:
        :return:
        """
        raise NotImplementedError

    def map_database(
        self, source_database: Any, target_database: Optional[Any] = None, **kwargs
    ) -> OBJECT_TYPE:
        """
        Transform source resource.

        :param source_database:
        :param target_database:
        :param kwargs:
        :return:
        """
        raise NotImplementedError

    def load_source_schema(self, path: Union[str, Path, dict]):
        """
        Sets source_schemaview from a schema path.

        :param path:
        :return:
        """
        if isinstance(path, Path):
            path = str(path)
        self.source_schemaview = SchemaView(path)

    def load_transformer_specification(self, path: Union[str, Path]):
        """
        Sets specification from a schema path.

        :param path:
        :return:
        """
        # self.specification = yaml_loader.load(str(path), TransformationSpecification)
        with open(path) as f:
            obj = yaml.safe_load(f)
            # necessary to expand first
            normalizer = ReferenceValidator(
                package_schemaview("linkml_transformer.datamodel.transformer_model")
            )
            normalizer.expand_all = True
            obj = normalizer.normalize(obj)
            self.specification = TransformationSpecification(**obj)

    def create_transformer_specification(self, obj: Dict[str, Any]):
        """
        Creates specification from a dict.

        TODO: this will no longer be necessary when pydantic supports inlined as dict

        :param path:
        :return:
        """
        normalizer = ReferenceValidator(
            package_schemaview("linkml_transformer.datamodel.transformer_model")
        )
        normalizer.expand_all = True
        obj = normalizer.normalize(obj)
        self.specification = TransformationSpecification(**obj)

    @property
    def derived_specification(self) -> Optional[TransformationSpecification]:
        if self._derived_specification is None:
            if self.specification is None:
                return None
            self._derived_specification = deepcopy(self.specification)
            induce_missing_values(self._derived_specification, self.source_schemaview)
        return self._derived_specification

    def _get_class_derivation(self, target_class_name: str) -> ClassDerivation:
        spec = self.derived_specification
        matching_tgt_class_derivs = [
            deriv
            for deriv in spec.class_derivations.values()
            if deriv.populated_from == target_class_name
            or (not deriv.populated_from and target_class_name == deriv.name)
        ]
        logger.debug(f"Target class derivations={matching_tgt_class_derivs}")
        if len(matching_tgt_class_derivs) != 1:
            raise ValueError(
                f"Could not find class derivation for {target_class_name} (results={len(matching_tgt_class_derivs)})"
            )
        cd = matching_tgt_class_derivs[0]
        ancmap = self._class_derivation_ancestors(cd)
        if ancmap:
            cd = deepcopy(cd)
            for anc in ancmap.values():
                for k, v in anc.__dict__.items():
                    if v is not None and v != []:
                        curr_v = getattr(cd, k, None)
                        if isinstance(curr_v, list):
                            curr_v.extend(v)
                        elif isinstance(curr_v, dict):
                            curr_v.update({**v, **curr_v})
                        else:
                            if curr_v is None:
                                setattr(cd, k, v)
        return cd

    def _class_derivation_ancestors(self, cd: ClassDerivation) -> Dict[str, ClassDerivation]:
        """
        Returns a map of all class derivations that are ancestors of the given class derivation.
        :param cd:
        :return:
        """
        spec = self.specification
        ancestors = {}
        parents = cd.mixins + ([cd.is_a] if cd.is_a else [])
        for parent in parents:
            ancestors[parent] = spec.class_derivations[parent]
            ancestors.update(self._class_derivation_ancestors(spec.class_derivations[parent]))
        return ancestors

    def _get_enum_derivation(self, target_enum_name: str) -> EnumDerivation:
        spec = self.derived_specification
        matching_tgt_enum_derivs = [
            deriv
            for deriv in spec.enum_derivations.values()
            if deriv.populated_from == target_enum_name
            or (not deriv.populated_from and target_enum_name == deriv.name)
        ]
        logger.debug(f"Target enum derivations={matching_tgt_enum_derivs}")
        if len(matching_tgt_enum_derivs) != 1:
            raise ValueError(f"Could not find what to derive from a source {target_enum_name}")
        return matching_tgt_enum_derivs[0]

    def _is_coerce_to_multivalued(
        self, slot_derivation: SlotDerivation, class_derivation: ClassDerivation
    ):
        cast_as = slot_derivation.cast_collection_as
        if cast_as and cast_as in [
            CollectionType.MultiValued,
            CollectionType.MultiValuedDict,
            CollectionType.MultiValuedDict,
        ]:
            return True
        if slot_derivation.stringification and slot_derivation.stringification.reversed:
            return True
        sv = self.target_schemaview
        if sv:
            slot = sv.induced_slot(slot_derivation.name, class_derivation.name)
            if slot.multivalued:
                return True
        return False

    def _is_coerce_to_singlevalued(
        self, slot_derivation: SlotDerivation, class_derivation: ClassDerivation
    ):
        cast_as = slot_derivation.cast_collection_as
        if cast_as and cast_as == CollectionType(CollectionType.SingleValued):
            return True
        if slot_derivation.stringification and not slot_derivation.stringification.reversed:
            return True
        sv = self.target_schemaview
        if sv:
            slot = sv.induced_slot(slot_derivation.name, class_derivation.name)
            if not slot.multivalued:
                return True
        return False

    def _coerce_datatype(self, v: Any, target_range: Optional[str]) -> Any:
        if target_range is None:
            return v
        if isinstance(v, list):
            return [self._coerce_datatype(v1, target_range) for v1 in v]
        if isinstance(v, dict):
            return {k: self._coerce_datatype(v1, target_range) for k, v1 in v.items()}
        cmap = {
            "integer": int,
            "float": float,
            "string": str,
            "boolean": bool,
        }
        cls = cmap.get(target_range, None)
        if not cls:
            logger.warning(f"Unknown target range {target_range}")
            return v
        if isinstance(v, cls):
            return v
        return cls(v)

    @property
    def curie_converter(self) -> Converter:
        if not self._curie_converter:
            self._curie_converter = Converter([])
            for prefix in self.source_schemaview.schema.prefixes.values():
                self._curie_converter.add_prefix(prefix.prefix_prefix, prefix.prefix_reference)
            for prefix in self.specification.prefixes.values():
                self._curie_converter.add_prefix(prefix.key, prefix.value)
        return self._curie_converter

    def expand_curie(self, curie: str) -> str:
        return self.curie_converter.expand(curie)

    def compress_uri(self, uri: str) -> str:
        return self.curie_converter.compress(uri)



"""
Basic unit conversion functions.

Currently only native pint units or UCUM units are supported.

For UCUM, the ucumvert library is used to convert UCUM units to pint units,
see `<https://github.com/dalito/ucumvert>`_.
"""

from enum import Enum
from functools import lru_cache
from typing import Any, Optional

import lark
import pint
from ucumvert import PintUcumRegistry


class UnitSystem(str, Enum):
    """Enumeration of supported unit systems."""

    UCUM = "ucum"
    IEC61360 = "iec61360"
    SI = "SI"


REGISTRIES = {
    UnitSystem.UCUM: PintUcumRegistry(),
}


class UndefinedUnitError(Exception):
    """
    Raised when a unit is not defined.

    Note: equivalent to the pint error, but the
    pint dependency is optional
    """


class QuantitySyntaxError(Exception):
    """
    Raised when a quantity cannot be parsed.
    """


class DimensionalityError(Exception):
    """
    Raised when a unit conversion cannot be performed.

    Note: equivalent to the pint error, but the
    pint dependency is optional
    """


def convert_units(
    magnitude: float, from_unit: str, to_unit: str, system: Optional[UnitSystem] = None
):
    """
    Convert a quantity between units.

    >>> convert_units(1, "m", "cm")
    100.0
    >>> convert_units(1, "m", "cm", system=UnitSystem.UCUM)
    100.0
    >>> convert_units(1, "m", "cm", system=UnitSystem.SI)
    100.0
    >>> convert_units(1.0, "hectare", "m^2", system=UnitSystem.SI)
    10000.0
    >>> convert_units(1.0, "hectare", "m ** 2", system=UnitSystem.SI)
    10000.0
    >>> convert_units(1.0, "hectare", "m ** 2", system=None)
    10000.0
    >>> convert_units(1.0, "km2", "m2", system=UnitSystem.UCUM)
    1000000.0

    :param magnitude:
    :param from_unit:
    :param to_unit:
    :return: converted magnitude
    """
    import pint

    ureg: pint.UnitRegistry = get_unit_registry(system)
    from_unit = normalize_unit(from_unit, system)
    to_unit = normalize_unit(to_unit, system)
    try:
        from_unit_q = ureg.parse_units(from_unit)
    except lark.exceptions.UnexpectedCharacters:
        raise UndefinedUnitError(f"Cannot parse unit: {from_unit}")
    except pint.errors.UndefinedUnitError:
        raise UndefinedUnitError(f"Unknown source unit: {from_unit}")
    quantity = magnitude * from_unit_q
    try:
        return quantity.to(to_unit).magnitude
    except pint.errors.UndefinedUnitError:
        raise UndefinedUnitError(f"Unknown target unit: {from_unit}")
    except pint.errors.DimensionalityError:
        raise DimensionalityError(f"Cannot convert from {from_unit} to {to_unit}")


@lru_cache()
def get_unit_registry(system: Optional[UnitSystem] = None) -> Any:
    """
    Get a unit registry.

    >>> ureg = get_unit_registry(UnitSystem.UCUM)
    >>> ureg.from_ucum("m/s2.kg")
    <Quantity(1.0, 'kilogram * meter / second ** 2')>
    >>> str(_)
    '1.0 kilogram * meter / second ** 2'
    >>> ureg.from_ucum("m[H2O]{35Cel}")  # UCUM code with annotation
    <Quantity(1, 'm_H2O')>
    >>> _.to("mbar")
    <Quantity(98.0665, 'millibar')>
    >>> ureg("degC")   # a standard pint unit
    <Quantity(1, 'degree_Celsius')>
    >>> ureg.from_ucum("g/m2")
    <Quantity(1.0, 'gram / meter ** 2')>
    >>> _.to(ureg.from_ucum("kg/m2"))
    <Quantity(0.001, 'kilogram / meter ** 2')>
    >>> ureg.from_ucum("nmol/mmol{Cre}")
    <Quantity(1.0, 'nanomole / millimole')>
    >>> sireg = get_unit_registry(UnitSystem.SI)
    >>> sireg("degC")
    <Quantity(1, 'degree_Celsius')>
    >>> sireg("ha")
    <Quantity(1, 'hectare')>

    :param system: currently only supported values are None or UnitSystem.UCUM
    :return:
    """
    import pint

    ureg = pint.UnitRegistry()
    if not system:
        return ureg
    if system in REGISTRIES:
        return REGISTRIES[system]
    if system.value in dir(ureg.sys):
        ureg.default_system = system.value
        return ureg
    raise NotImplementedError(f"Unknown unit system: {system}")


def normalize_unit(unit: str, system: Optional[UnitSystem] = None) -> str:
    if system is None:
        return unit
    if system == UnitSystem.UCUM:
        try:
            return str(get_unit_registry(system).from_ucum(unit))
        except pint.errors.UndefinedUnitError:
            raise UndefinedUnitError(f"Unknown unit: {unit}")
        except lark.exceptions.UnexpectedCharacters:
            raise UndefinedUnitError(f"Cannot parse unit: {unit}")
    else:
        return unit



import json
import logging
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Type, Union

import yaml
from asteval import Interpreter
from linkml_runtime import SchemaView
from linkml_runtime.index.object_index import ObjectIndex
from linkml_runtime.utils.yamlutils import YAMLRoot
from pydantic import BaseModel

from linkml_transformer.datamodel.transformer_model import (
    CollectionType,
    SerializationSyntaxType,
    SlotDerivation,
)
from linkml_transformer.functions.unit_conversion import UnitSystem, convert_units
from linkml_transformer.transformer.transformer import OBJECT_TYPE, Transformer
from linkml_transformer.utils.dynamic_object import dynamic_object
from linkml_transformer.utils.eval_utils import eval_expr

DICT_OBJ = Dict[str, Any]


logger = logging.getLogger(__name__)


@dataclass
class ObjectTransformer(Transformer):
    """
    A Transformer that works on in-memory dict objects.

    This works recursively
    """

    object_index: ObjectIndex = None

    def index(self, source_obj: Any, target: str = None):
        """
        Create an index over a container object.

        :param source_obj: source data structure to be indexed
        :param target: class to convert source object into
        """
        if isinstance(source_obj, dict):
            if target is None:
                [target] = [
                    c.name for c in self.source_schemaview.all_classes().values() if c.tree_root
                ]
            if target is None:
                raise ValueError(f"target must be passed if source_obj is dict: {source_obj}")
            source_obj_typed = dynamic_object(source_obj, self.source_schemaview, target)
            self.object_index = ObjectIndex(source_obj_typed, schemaview=self.source_schemaview)
        else:
            self.object_index = ObjectIndex(source_obj, schemaview=self.source_schemaview)

    def map_object(
        self,
        source_obj: OBJECT_TYPE,
        source_type: str = None,
        target_type: str = None,
    ) -> Union[DICT_OBJ, Any]:
        """
        Transform a source object into a target object.

        :param source_obj: source data structure
        :param source_type: source_obj instantiates this (may be class, type, or enum)
        :param target_type: target_obj instantiates this (may be class, type, or enum)
        :return: transformed data, either as type target_type or a dictionary
        """
        sv = self.source_schemaview
        if source_type is None and sv is None:
            # TODO: use smarter method
            source_type = list(self.specification.class_derivations.values())[0].name
        if source_type is None and sv is not None:
            source_types = [c.name for c in sv.all_classes().values() if c.tree_root]
            if len(source_types) == 1:
                source_type = source_types[0]
            elif len(source_types) > 1:
                raise ValueError("No source type specified and multiple root classes found")
            elif len(source_types) == 0:
                if len(sv.all_classes()) == 1:
                    source_type = list(sv.all_classes().keys())[0]
                else:
                    raise ValueError("No source type specified and no root classes found")

        if source_type in sv.all_types():
            if target_type:
                if target_type == "string":
                    return str(source_obj)
                elif target_type == "integer":
                    return int(source_obj)
                elif target_type == "float" or target_type == "double":
                    return float(source_obj)
                elif target_type == "uri":
                    return self.expand_curie(source_obj)
                elif target_type == "curie":
                    return self.compress_uri(source_obj)
            return source_obj
        if source_type in sv.all_enums():
            # TODO: enum derivations
            return self.transform_enum(source_obj, source_type, source_obj)
            # return str(source_obj)
        source_obj_typed = None
        if isinstance(source_obj, (BaseModel, YAMLRoot)):
            # ensure dict
            source_obj_typed = source_obj
            source_obj = vars(source_obj)
        if not isinstance(source_obj, dict):
            logger.warning(f"Unexpected: {source_obj} for type {source_type}")
            return source_obj
        class_deriv = self._get_class_derivation(source_type)
        tgt_attrs = {}
        # map each slot assignment in source_obj, if there is a slot_derivation
        for slot_derivation in class_deriv.slot_derivations.values():
            v = None
            source_class_slot = None
            if slot_derivation.unit_conversion:
                v = self._perform_unit_conversion(slot_derivation, source_obj, sv, source_type)
            elif slot_derivation.expr:
                if self.object_index:
                    if not source_obj_typed:
                        source_obj_dyn = dynamic_object(source_obj, sv, source_type)
                    else:
                        source_obj_dyn = source_obj_typed
                    ctxt_obj = self.object_index.bless(source_obj_dyn)
                    ctxt_dict = {
                        k: getattr(ctxt_obj, k)
                        for k in ctxt_obj._attributes()
                        if not k.startswith("_")
                    }
                else:
                    do = dynamic_object(source_obj, sv, source_type)
                    ctxt_obj = do
                    ctxt_dict = vars(do)

                try:
                    v = eval_expr(slot_derivation.expr, **ctxt_dict, NULL=None)
                except Exception:
                    if not self.unrestricted_eval:
                        raise RuntimeError(f"Expression not in safe subset: {slot_derivation.expr}")
                    aeval = Interpreter(usersyms={"src": ctxt_obj, "target": None})
                    aeval(slot_derivation.expr)
                    v = aeval.symtable["target"]
            elif slot_derivation.populated_from:
                v = source_obj.get(slot_derivation.populated_from, None)
                source_class_slot = sv.induced_slot(slot_derivation.populated_from, source_type)
                logger.debug(
                    f"Pop slot {slot_derivation.name} => {v} using {slot_derivation.populated_from} // {source_obj}"
                )
            elif slot_derivation.sources:
                vmap = {s: source_obj.get(s, None) for s in slot_derivation.sources}
                vmap = {k: v for k, v in vmap.items() if v is not None}
                if len(vmap.keys()) > 1:
                    raise ValueError(f"Multiple sources for {slot_derivation.name}: {vmap}")
                elif len(vmap.keys()) == 1:
                    v = list(vmap.values())[0]
                    source_class_slot_name = list(vmap.keys())[0]
                    source_class_slot = sv.induced_slot(source_class_slot_name, source_type)
                else:
                    v = None
                    source_class_slot = None

                logger.debug(
                    f"Pop slot {slot_derivation.name} => {v} using {slot_derivation.populated_from} // {source_obj}"
                )
            else:
                source_class_slot = sv.induced_slot(slot_derivation.name, source_type)
                v = source_obj.get(slot_derivation.name, None)
            if source_class_slot and v is not None:
                # slot is mapped and there is a value in the assignment
                target_range = slot_derivation.range
                source_class_slot_range = source_class_slot.range
                if source_class_slot.multivalued:
                    if isinstance(v, list):
                        v = [self.map_object(v1, source_class_slot_range, target_range) for v1 in v]
                    elif isinstance(v, dict):
                        v = {
                            k1: self.map_object(v1, source_class_slot_range, target_range)
                            for k1, v1 in v.items()
                        }
                    else:
                        v = [v]
                else:
                    v = self.map_object(v, source_class_slot_range, target_range)
                if (
                    self._is_coerce_to_multivalued(slot_derivation, class_deriv)
                    and v is not None
                    and not isinstance(v, list)
                ):
                    v = self._singlevalued_to_multivalued(v, slot_derivation)
                if self._is_coerce_to_singlevalued(slot_derivation, class_deriv) and isinstance(
                    v, list
                ):
                    v = self._multivalued_to_singlevalued(v, slot_derivation)
                v = self._coerce_datatype(v, target_range)
                if slot_derivation.dictionary_key and isinstance(v, list):
                    # List to CompactDict
                    v = {v1[slot_derivation.dictionary_key]: v1 for v1 in v}
                    for v1 in v.values():
                        del v1[slot_derivation.dictionary_key]
                elif (
                    slot_derivation.cast_collection_as
                    and slot_derivation.cast_collection_as == CollectionType.MultiValuedList
                    and isinstance(v, dict)
                ):
                    # CompactDict to List
                    src_rng = source_class_slot.range
                    src_rng_id_slot = self.source_schemaview.get_identifier_slot(
                        src_rng, use_key=True
                    )
                    if src_rng_id_slot:
                        v = [{**v1, src_rng_id_slot.name: k} for k, v1 in v.items()]
                    else:
                        v = list(v.values())
            tgt_attrs[str(slot_derivation.name)] = v
        return tgt_attrs

    def _perform_unit_conversion(
        self, slot_derivation: SlotDerivation, source_obj: Any, sv: SchemaView, source_type: str
    ) -> Union[float, Dict]:
        uc = slot_derivation.unit_conversion
        curr_v = source_obj.get(slot_derivation.populated_from, None)
        system = UnitSystem.UCUM
        if curr_v is not None:
            slot = sv.induced_slot(slot_derivation.populated_from, source_type)
            to_unit = uc.target_unit
            if uc.source_unit_slot:
                from_unit = curr_v.get(uc.source_unit_slot, None)
                if from_unit is None:
                    raise ValueError(
                        f"Could not determine unit from {curr_v}" f" using {uc.source_unit_slot}"
                    )
                magnitude = curr_v.get(uc.source_magnitude_slot, None)
                if magnitude is None:
                    raise ValueError(
                        f"Could not determine magnitude from {curr_v}"
                        f" using {uc.source_magnitude_slot}"
                    )
            else:
                if slot.unit.ucum_code:
                    from_unit = slot.unit.ucum_code
                elif slot.unit.iec61360code:
                    from_unit = slot.unit.iec61360code
                    system = UnitSystem.IEC61360
                else:
                    system = None
                    if slot.unit.symbol:
                        from_unit = slot.unit.symbol
                    elif slot.unit.abbreviation:
                        from_unit = slot.unit.abbreviation
                    elif slot.unit.descriptive_name:
                        from_unit = slot.unit.descriptive_name
                    else:
                        raise NotImplementedError(f"Cannot determine unit system for {slot.unit}")
                magnitude = curr_v
            if not from_unit:
                raise ValueError(f"Could not determine from_unit for {slot_derivation}")
            if not to_unit:
                to_unit = from_unit
                # raise ValueError(f"Could not determine to_unit for {slot_derivation}")
            if from_unit == to_unit:
                v = magnitude
            else:
                v = convert_units(
                    magnitude,
                    from_unit=from_unit,
                    to_unit=to_unit,
                    system=system,
                )
            if uc.target_magnitude_slot:
                v = {uc.target_magnitude_slot: v, uc.target_unit_slot: to_unit}
            return v

    def _multivalued_to_singlevalued(self, vs: List[Any], slot_derivation: SlotDerivation) -> Any:
        if slot_derivation.stringification:
            stringification = slot_derivation.stringification
            delimiter = stringification.delimiter
            if delimiter:
                return delimiter.join(vs)
            elif stringification.syntax:
                if stringification.syntax == SerializationSyntaxType.JSON:
                    return json.dumps(vs)
                elif stringification.syntax == SerializationSyntaxType.YAML:
                    return yaml.dump(vs, default_flow_style=True).strip()
                else:
                    raise ValueError(f"Unknown syntax: {stringification.syntax}")
            else:
                raise ValueError(f"Cannot convert multivalued to single valued: {vs}; no delimiter")
        if len(vs) > 1:
            raise ValueError(f"Cannot coerce multiple values {vs}")
        if len(vs) == 0:
            return None
        else:
            return vs[0]

    def _singlevalued_to_multivalued(self, v: Any, slot_derivation: SlotDerivation) -> List[Any]:
        stringification = slot_derivation.stringification
        if stringification:
            delimiter = stringification.delimiter
            if delimiter:
                vs = v.split(slot_derivation.stringification.delimiter)
                if vs == [""]:
                    vs = []
            elif stringification.syntax:
                syntax = stringification.syntax
                if syntax == SerializationSyntaxType.JSON:
                    vs = json.loads(v)
                elif syntax == SerializationSyntaxType.YAML:
                    vs = yaml.safe_load(v)
                else:
                    raise ValueError(f"Unknown syntax: {syntax}")
            else:
                raise ValueError(f"Cannot convert single valued to multivalued: {v}; no delimiter")
            return vs
        return [v]

    def transform_object(
        self,
        source_obj: Union[YAMLRoot, BaseModel],
        target_class: Optional[Union[Type[YAMLRoot], Type[BaseModel]]] = None,
    ) -> Union[YAMLRoot, BaseModel]:
        """
        Transform an object into an object of class target_class.

        :param source_obj: source object
        :type source_obj: Union[YAMLRoot, BaseModel]
        :param target_class: class to transform the object into, defaults to None
        :type target_class: Optional[Union[Type[YAMLRoot], Type[BaseModel]]], optional
        :return: transformed object of class target_class
        :rtype: Union[YAMLRoot, BaseModel]
        """
        if not target_class:
            raise ValueError("No target_class specified for transform_object")

        source_type = type(source_obj)
        source_type_name = source_type.__name__
        # if isinstance(source_obj, YAMLRoot):
        #    source_obj_dict = json_dumper.to_dict(source_obj)
        # elif isinstance(source_obj, BaseModel):
        #    source_obj_dict = source_obj.dict()
        # else:
        #    raise ValueError(f"Do not know how to handle type: {typ}")
        tr_obj_dict = self.map_object(source_obj, source_type_name)
        return target_class(**tr_obj_dict)

    def transform_enum(self, source_value: str, enum_name: str, source_obj: Any) -> Optional[str]:
        enum_deriv = self._get_enum_derivation(enum_name)
        if enum_deriv.expr:
            try:
                if enum_deriv.expr:
                    v = eval_expr(enum_deriv.expr, **source_obj, NULL=None)
            except Exception:
                aeval = Interpreter(usersyms={"src": source_obj, "target": None})
                aeval(enum_deriv.expr)
                v = aeval.symtable["target"]
            if v is not None:
                return v
        for pv_deriv in enum_deriv.permissible_value_derivations.values():
            if source_value == pv_deriv.populated_from:
                return pv_deriv.name
            if source_value in pv_deriv.sources:
                return pv_deriv.name
        if enum_deriv.mirror_source:
            return str(source_value)
        else:
            return None



"""
Importers are responsible for translating an external mapping specification into the LinkML transformation model.

For example:

 - SSSOM to LinkML-Transformer Specifications
 - YARRRML to LinkML-Transformer Specifications
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any

from linkml_transformer.datamodel.transformer_model import TransformationSpecification


@dataclass
class Importer(ABC):
    """
    Base class for all importers.

    A compiler will translate an external mapping spec into the LinkML transformation model.

    An example importer would be sssom.
    """

    @abstractmethod
    def import_specification(self, input: Any) -> TransformationSpecification:
        """
        Import into a specification

        :param specification:
        :return:
        """
        raise NotImplementedError



"""
A SchemaMapper translates a source schema and transformation specification into a target schema.

AKA profiling
"""

import logging
from collections import defaultdict
from copy import copy
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Tuple

from linkml_runtime import SchemaView
from linkml_runtime.dumpers import json_dumper
from linkml_runtime.linkml_model import (
    ClassDefinition,
    ClassDefinitionName,
    Element,
    EnumDefinition,
    PermissibleValue,
    SchemaDefinition,
    SlotDefinition,
)
from linkml_runtime.linkml_model.units import UnitOfMeasure

from linkml_transformer.datamodel.transformer_model import (
    ClassDerivation,
    CollectionType,
    CopyDirective,
    EnumDerivation,
    TransformationSpecification,
)
from linkml_transformer.transformer.transformer import Transformer

logger = logging.getLogger(__name__)


@dataclass
class SchemaMapper:
    """
    Translates a source schema and transformation specification into a target schema.
    """

    source_schemaview: SchemaView = None

    transformer: Transformer = None

    source_to_target_class_mappings: Dict[str, List[str]] = field(
        default_factory=lambda: defaultdict(list)
    )

    slot_info: Dict[Tuple[str, str], Any] = field(default_factory=lambda: {})

    def derive_schema(
        self,
        specification: Optional[TransformationSpecification] = None,
        target_schema_id: Optional[str] = None,
        target_schema_name: Optional[str] = None,
        suffix="-derived",
    ) -> SchemaDefinition:
        """
        Use a transformation specification to generate a target/profile schema from a source schema.

        :param specification:
        :return:
        """
        if specification is None:
            specification = self.transformer.specification
        source_schemaview = self.source_schemaview
        source_schema = source_schemaview.schema
        if target_schema_id is None:
            target_schema_id = source_schema.id + suffix
        if target_schema_name is None:
            target_schema_name = source_schema.name + suffix
        target_schema = SchemaDefinition(id=target_schema_id, name=target_schema_name)
        for im in source_schema.imports:
            target_schema.imports.append(im)
        for prefix in source_schema.prefixes.values():
            target_schema.prefixes[prefix.prefix_prefix] = prefix
        for class_derivation in specification.class_derivations.values():
            class_definition = self._derive_class(class_derivation)
            target_schema.classes[class_definition.name] = class_definition
        for enum_derivation in specification.enum_derivations.values():
            enum_definition = self._derive_enum(enum_derivation)
            target_schema.enums[enum_definition.name] = enum_definition
        target_schema.default_range = source_schema.default_range
        for cd in target_schema.classes.values():
            self._rewire_class(cd)
        for (cn, sn), info in self.slot_info.items():
            cd = target_schema.classes[cn]
            sd = cd.attributes[sn]
            for k, v in info.items():
                setattr(sd, k, v)
        return target_schema

    def _derive_class(self, class_derivation: ClassDerivation) -> ClassDefinition:
        """
        Derive a class from a class derivation.
        """
        populated_from = class_derivation.populated_from
        if not populated_from:
            populated_from = class_derivation.name
        logger.info(f"Populating {class_derivation.name} from {populated_from}")
        source_class = self.source_schemaview.get_class(populated_from)
        if source_class is None:
            logger.warning(f"No such class {populated_from}")
            target_class = ClassDefinition(name=class_derivation.name)
        else:
            target_class = copy(source_class)
            target_class.from_schema = None
            target_class.name = class_derivation.name
            target_class.slots = []
            target_class.attributes = {}
            target_class.slot_usage = {}
        for slot_derivation in class_derivation.slot_derivations.values():
            slot_definition = self._derive_slot(slot_derivation)
            target_class.attributes[slot_definition.name] = slot_definition
        self.source_to_target_class_mappings[populated_from].append(target_class.name)
        if class_derivation.overrides:
            curr = json_dumper.to_dict(target_class)
            for k, v in class_derivation.overrides.items():
                curr[k] = v
            target_class = ClassDefinition(**curr)
        return target_class

    def _derive_enum(self, enum_derivation: EnumDerivation) -> EnumDefinition:
        """
        Derive an enum from an enum derivation.

        :param enum_derivation:
        :return:
        """
        populated_from = enum_derivation.populated_from
        if not populated_from:
            populated_from = enum_derivation.name
        source_enum = self.source_schemaview.get_enum(populated_from)
        if source_enum is None:
            logger.warning(f"No such enum {populated_from}")
            target_enum = ClassDefinition(name=enum_derivation.name)
        else:
            target_enum = copy(source_enum)
            target_enum.from_schema = None
            target_enum.name = enum_derivation.name
            target_enum.slots = []
            target_enum.attributes = {}
            target_enum.slot_usage = {}
        for pv_derivation in enum_derivation.permissible_value_derivations.values():
            if pv_derivation.populated_from:
                pv = PermissibleValue(text=pv_derivation.populated_from)
                target_enum.permissible_values[pv.text] = pv
            elif pv_derivation.sources:
                for source in pv_derivation.sources:
                    pv = PermissibleValue(text=source)
                    target_enum.permissible_values[pv.text] = pv
            else:
                raise ValueError(f"Missing populated_from or sources for {pv_derivation}")
        if enum_derivation.mirror_source:
            for pv in source_enum.permissible_values.values():
                if pv.text not in target_enum.permissible_values:
                    target_enum.permissible_values[pv.text] = copy(pv)
        self.source_to_target_class_mappings[populated_from].append(target_enum.name)
        return target_enum

    def _derive_slot(self, slot_derivation) -> SlotDefinition:
        """
        Derive a slot from a slot derivation.
        """
        populated_from = slot_derivation.populated_from
        if not populated_from:
            populated_from = slot_derivation.name
        source_slot = self.source_schemaview.get_slot(populated_from)
        if source_slot is None:
            target_slot = SlotDefinition(name=slot_derivation.name)
        else:
            target_slot = copy(source_slot)
            target_slot.from_schema = None
            target_slot.owner = None
            target_slot.name = slot_derivation.name
        if slot_derivation.range:
            target_slot.range = slot_derivation.range
        if slot_derivation.unit_conversion:
            target_slot.unit = UnitOfMeasure(ucum_code=slot_derivation.unit_conversion.target_unit)
        if slot_derivation.stringification:
            if slot_derivation.stringification.reversed:
                target_slot.multivalued = True
            else:
                target_slot.multivalued = False
        if slot_derivation.dictionary_key:
            target_slot.inlined = True
            target_slot.inlined_as_list = False
            self.slot_info[(target_slot.range, slot_derivation.dictionary_key)] = {
                "identifier": True
            }
        if slot_derivation.cast_collection_as:
            if slot_derivation.cast_collection_as == CollectionType.MultiValued:
                target_slot.inlined = True
            elif slot_derivation.cast_collection_as == CollectionType.MultiValuedList:
                target_slot.inlined_as_list = True
            elif slot_derivation.cast_collection_as == CollectionType.MultiValuedDict:
                target_slot.inlined = True
                target_slot.inlined_as_list = False
        return target_slot

    def _rewire_class(self, class_definition: ClassDefinition):
        if class_definition.is_a:
            class_definition.is_a = self._rewire_parent(class_definition, class_definition.is_a)
        mixins = [self._rewire_parent(class_definition, m) for m in class_definition.mixins]
        class_definition.mixins = [m for m in mixins if m is not None]

    def _rewire_parent(
        self, class_definition: ClassDefinition, parent: ClassDefinitionName
    ) -> Optional[str]:
        if parent in self.source_to_target_class_mappings:
            new_parents = self.source_to_target_class_mappings[parent]
            if len(new_parents) > 1:
                raise ValueError(
                    f"Cannot rewire to non-isomorphic mappings {parent} => {new_parents}"
                )
            if len(new_parents) == 1:
                return new_parents[0]
        parent_cls = self.source_schemaview.get_class(parent)
        if parent_cls.is_a:
            return self._rewire_parent(class_definition, parent_cls.is_a)
        return None

    def copy_attributes(
        self,
        target_element: Element,
        source_element: Element,
        copy_directive: CopyDirective,
    ) -> None:
        """
        Copy attributes from source to target according to a directive.

        :param target_element:
        :param source_element:
        :param copy_directive:
        :return:
        """
        for k, v in vars(source_element).items():
            included = False
            if copy_directive.include_all:
                included = True
            if k in copy_directive.include:
                included = True
            if k in copy_directive.exclude:
                included = False
            if included:
                setattr(target_element, k, v)



from __future__ import annotations

import re
import sys
from datetime import date, datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel as BaseModel
from pydantic import ConfigDict, Field, field_validator

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


metamodel_version = "None"
version = "None"


class ConfiguredBaseModel(BaseModel):
    model_config = ConfigDict(
        validate_assignment=True,
        validate_default=True,
        extra="forbid",
        arbitrary_types_allowed=True,
        use_enum_values=True,
    )


class CollectionType(str, Enum):
    SingleValued = "SingleValued"

    MultiValued = "MultiValued"

    MultiValuedList = "MultiValuedList"

    MultiValuedDict = "MultiValuedDict"


class SerializationSyntaxType(str, Enum):
    JSON = "JSON"

    YAML = "YAML"

    TURTLE = "TURTLE"


class SpecificationComponent(ConfiguredBaseModel):
    description: Optional[str] = Field(
        None, description="""description of the specification component"""
    )
    implements: Optional[List[str]] = Field(
        default_factory=list,
        description="""A reference to a specification that this component implements.""",
    )
    comments: Optional[List[str]] = Field(
        default_factory=list,
        description="""A list of comments about this component. Comments are free text, and may be used to provide additional information about the component, including instructions for its use.""",
    )


class TransformationSpecification(SpecificationComponent):
    """
    A collection of mappings between source and target classes
    """

    id: Optional[str] = Field(
        None, description="""Unique identifier for this transformation specification"""
    )
    title: Optional[str] = Field(
        None, description="""human readable title for this transformation specification"""
    )
    prefixes: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict, description="""maps prefixes to URL expansions"""
    )
    source_schema: Optional[str] = Field(
        None, description="""name of the schema that describes the source (input) objects"""
    )
    target_schema: Optional[str] = Field(
        None, description="""name of the schema that describes the target (output) objects"""
    )
    class_derivations: Optional[Dict[str, ClassDerivation]] = Field(
        default_factory=dict,
        description="""Instructions on how to derive a set of classes in the target schema from classes in the source schema.""",
    )
    enum_derivations: Optional[Dict[str, EnumDerivation]] = Field(
        default_factory=dict,
        description="""Instructions on how to derive a set of enums in the target schema""",
    )
    slot_derivations: Optional[Dict[str, SlotDerivation]] = Field(
        default_factory=dict,
        description="""Instructions on how to derive a set of top level slots in the target schema""",
    )
    description: Optional[str] = Field(
        None, description="""description of the specification component"""
    )
    implements: Optional[List[str]] = Field(
        default_factory=list,
        description="""A reference to a specification that this component implements.""",
    )
    comments: Optional[List[str]] = Field(
        default_factory=list,
        description="""A list of comments about this component. Comments are free text, and may be used to provide additional information about the component, including instructions for its use.""",
    )


class ElementDerivation(SpecificationComponent):
    """
    An abstract grouping for classes that provide a specification of how to  derive a target element from a source element.
    """

    name: str = Field(..., description="""Name of the element in the target schema""")
    copy_directives: Optional[Dict[str, CopyDirective]] = Field(default_factory=dict)
    overrides: Optional[Any] = Field(None, description="""overrides source schema slots""")
    is_a: Optional[str] = Field(None)
    mixins: Optional[List[str]] = Field(default_factory=list)
    value_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table that is applied directly to mappings, in order of precedence""",
    )
    expression_to_value_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict, description="""A mapping table in which the keys are expressions"""
    )
    expression_to_expression_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table in which the keys and values are expressions""",
    )
    mirror_source: Optional[bool] = Field(None)
    description: Optional[str] = Field(
        None, description="""description of the specification component"""
    )
    implements: Optional[List[str]] = Field(
        default_factory=list,
        description="""A reference to a specification that this component implements.""",
    )
    comments: Optional[List[str]] = Field(
        default_factory=list,
        description="""A list of comments about this component. Comments are free text, and may be used to provide additional information about the component, including instructions for its use.""",
    )


class ClassDerivation(ElementDerivation):
    """
    A specification of how to derive a target class from a source class.
    """

    populated_from: Optional[str] = Field(
        None, description="""Name of the class in the source schema"""
    )
    sources: Optional[List[str]] = Field(default_factory=list)
    joins: Optional[Dict[str, AliasedClass]] = Field(
        default_factory=dict,
        description="""Additional classes to be joined to derive instances of the target class""",
    )
    slot_derivations: Optional[Dict[str, SlotDerivation]] = Field(default_factory=dict)
    name: str = Field(..., description="""Name of the element in the target schema""")
    copy_directives: Optional[Dict[str, CopyDirective]] = Field(default_factory=dict)
    overrides: Optional[Any] = Field(None, description="""overrides source schema slots""")
    is_a: Optional[str] = Field(None)
    mixins: Optional[List[str]] = Field(default_factory=list)
    value_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table that is applied directly to mappings, in order of precedence""",
    )
    expression_to_value_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict, description="""A mapping table in which the keys are expressions"""
    )
    expression_to_expression_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table in which the keys and values are expressions""",
    )
    mirror_source: Optional[bool] = Field(None)
    description: Optional[str] = Field(
        None, description="""description of the specification component"""
    )
    implements: Optional[List[str]] = Field(
        default_factory=list,
        description="""A reference to a specification that this component implements.""",
    )
    comments: Optional[List[str]] = Field(
        default_factory=list,
        description="""A list of comments about this component. Comments are free text, and may be used to provide additional information about the component, including instructions for its use.""",
    )


class AliasedClass(ConfiguredBaseModel):
    """
    alias-class key value pairs for classes
    """

    alias: str = Field(..., description="""name of the class to be aliased""")
    class_named: Optional[str] = Field(None, description="""local alias for the class""")


class SlotDerivation(ElementDerivation):
    """
    A specification of how to derive the value of a target slot from a source slot
    """

    name: str = Field(..., description="""Target slot name""")
    populated_from: Optional[str] = Field(None, description="""Source slot name""")
    sources: Optional[List[str]] = Field(default_factory=list)
    derived_from: Optional[List[str]] = Field(
        default_factory=list,
        description="""Source slots that are used to derive this slot. This can be computed from the expr, if the expr is declarative.""",
    )
    expr: Optional[str] = Field(
        None,
        description="""An expression to be evaluated on the source object to derive the target slot. Should be specified using the LinkML expression language.""",
    )
    range: Optional[str] = Field(None)
    unit_conversion: Optional[UnitConversionConfiguration] = Field(None)
    inverse_of: Optional[Inverse] = Field(
        None,
        description="""Used to specify a class-slot tuple that is the inverse of the derived/target slot. This is used primarily for mapping to relational databases or formalisms that do not allow multiple values. The class representing the repeated element has a foreign key slot inserted in that 'back references' the original multivalued slot.""",
    )
    hide: Optional[bool] = Field(None, description="""True if this is suppressed""")
    type_designator: Optional[bool] = Field(None)
    cast_collection_as: Optional[CollectionType] = Field(None)
    dictionary_key: Optional[str] = Field(None)
    stringification: Optional[StringificationConfiguration] = Field(None)
    copy_directives: Optional[Dict[str, CopyDirective]] = Field(default_factory=dict)
    overrides: Optional[Any] = Field(None, description="""overrides source schema slots""")
    is_a: Optional[str] = Field(None)
    mixins: Optional[List[str]] = Field(default_factory=list)
    value_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table that is applied directly to mappings, in order of precedence""",
    )
    expression_to_value_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict, description="""A mapping table in which the keys are expressions"""
    )
    expression_to_expression_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table in which the keys and values are expressions""",
    )
    mirror_source: Optional[bool] = Field(None)
    description: Optional[str] = Field(
        None, description="""description of the specification component"""
    )
    implements: Optional[List[str]] = Field(
        default_factory=list,
        description="""A reference to a specification that this component implements.""",
    )
    comments: Optional[List[str]] = Field(
        default_factory=list,
        description="""A list of comments about this component. Comments are free text, and may be used to provide additional information about the component, including instructions for its use.""",
    )


class EnumDerivation(ElementDerivation):
    """
    A specification of how to derive the value of a target enum from a source enum
    """

    name: str = Field(..., description="""Target enum name""")
    populated_from: Optional[str] = Field(None, description="""Source enum name""")
    sources: Optional[List[str]] = Field(default_factory=list)
    expr: Optional[str] = Field(
        None,
        description="""An expression to be evaluated on the source object to derive the target slot. Should be specified using the LinkML expression language.""",
    )
    hide: Optional[bool] = Field(None, description="""True if this is suppressed""")
    permissible_value_derivations: Optional[Dict[str, PermissibleValueDerivation]] = Field(
        default_factory=dict,
        description="""Instructions on how to derive a set of PVs in the target schema""",
    )
    copy_directives: Optional[Dict[str, CopyDirective]] = Field(default_factory=dict)
    overrides: Optional[Any] = Field(None, description="""overrides source schema slots""")
    is_a: Optional[str] = Field(None)
    mixins: Optional[List[str]] = Field(default_factory=list)
    value_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table that is applied directly to mappings, in order of precedence""",
    )
    expression_to_value_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict, description="""A mapping table in which the keys are expressions"""
    )
    expression_to_expression_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table in which the keys and values are expressions""",
    )
    mirror_source: Optional[bool] = Field(None)
    description: Optional[str] = Field(
        None, description="""description of the specification component"""
    )
    implements: Optional[List[str]] = Field(
        default_factory=list,
        description="""A reference to a specification that this component implements.""",
    )
    comments: Optional[List[str]] = Field(
        default_factory=list,
        description="""A list of comments about this component. Comments are free text, and may be used to provide additional information about the component, including instructions for its use.""",
    )


class PermissibleValueDerivation(ElementDerivation):
    """
    A specification of how to derive the value of a PV from a source enum
    """

    name: str = Field(..., description="""Target permissible value text""")
    expr: Optional[str] = Field(None)
    populated_from: Optional[str] = Field(None, description="""Source permissible value""")
    sources: Optional[List[str]] = Field(default_factory=list)
    hide: Optional[bool] = Field(None)
    copy_directives: Optional[Dict[str, CopyDirective]] = Field(default_factory=dict)
    overrides: Optional[Any] = Field(None, description="""overrides source schema slots""")
    is_a: Optional[str] = Field(None)
    mixins: Optional[List[str]] = Field(default_factory=list)
    value_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table that is applied directly to mappings, in order of precedence""",
    )
    expression_to_value_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict, description="""A mapping table in which the keys are expressions"""
    )
    expression_to_expression_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table in which the keys and values are expressions""",
    )
    mirror_source: Optional[bool] = Field(None)
    description: Optional[str] = Field(
        None, description="""description of the specification component"""
    )
    implements: Optional[List[str]] = Field(
        default_factory=list,
        description="""A reference to a specification that this component implements.""",
    )
    comments: Optional[List[str]] = Field(
        default_factory=list,
        description="""A list of comments about this component. Comments are free text, and may be used to provide additional information about the component, including instructions for its use.""",
    )


class PrefixDerivation(ElementDerivation):
    name: str = Field(..., description="""Name of the element in the target schema""")
    copy_directives: Optional[Dict[str, CopyDirective]] = Field(default_factory=dict)
    overrides: Optional[Any] = Field(None, description="""overrides source schema slots""")
    is_a: Optional[str] = Field(None)
    mixins: Optional[List[str]] = Field(default_factory=list)
    value_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table that is applied directly to mappings, in order of precedence""",
    )
    expression_to_value_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict, description="""A mapping table in which the keys are expressions"""
    )
    expression_to_expression_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table in which the keys and values are expressions""",
    )
    mirror_source: Optional[bool] = Field(None)
    description: Optional[str] = Field(
        None, description="""description of the specification component"""
    )
    implements: Optional[List[str]] = Field(
        default_factory=list,
        description="""A reference to a specification that this component implements.""",
    )
    comments: Optional[List[str]] = Field(
        default_factory=list,
        description="""A list of comments about this component. Comments are free text, and may be used to provide additional information about the component, including instructions for its use.""",
    )


class UnitConversionConfiguration(ConfiguredBaseModel):
    target_unit: Optional[str] = Field(None)
    target_unit_scheme: Optional[str] = Field(None)
    source_unit: Optional[str] = Field(None)
    source_unit_scheme: Optional[str] = Field(None)
    source_unit_slot: Optional[str] = Field(None)
    source_magnitude_slot: Optional[str] = Field(None)
    target_unit_slot: Optional[str] = Field(None)
    target_magnitude_slot: Optional[str] = Field(None)


class StringificationConfiguration(ConfiguredBaseModel):
    delimiter: Optional[str] = Field(None)
    reversed: Optional[bool] = Field(None)
    syntax: Optional[SerializationSyntaxType] = Field(None)


class Inverse(ConfiguredBaseModel):
    """
    Used for back references in mapping to relational model
    """

    slot_name: Optional[str] = Field(None)
    class_name: Optional[str] = Field(None)


class KeyVal(ConfiguredBaseModel):
    key: str = Field(...)
    value: Optional[Any] = Field(None)


class CopyDirective(ConfiguredBaseModel):
    """
    Instructs a Schema Mapper in how to map to a target schema. Not used for data transformation.
    """

    element_name: str = Field(...)
    copy_all: Optional[bool] = Field(None)
    exclude_all: Optional[bool] = Field(None)
    exclude: Optional[Any] = Field(None)
    include: Optional[Any] = Field(None)
    add: Optional[Any] = Field(None)


# Model rebuild
# see https://pydantic-docs.helpmanual.io/usage/models/#rebuilding-a-model
SpecificationComponent.model_rebuild()
TransformationSpecification.model_rebuild()
ElementDerivation.model_rebuild()
ClassDerivation.model_rebuild()
AliasedClass.model_rebuild()
SlotDerivation.model_rebuild()
EnumDerivation.model_rebuild()
PermissibleValueDerivation.model_rebuild()
PrefixDerivation.model_rebuild()
UnitConversionConfiguration.model_rebuild()
StringificationConfiguration.model_rebuild()
Inverse.model_rebuild()
KeyVal.model_rebuild()
CopyDirective.model_rebuild()



import logging
from dataclasses import dataclass
from typing import Any, Dict, Optional, Union

from duckdb import DuckDBPyConnection

from linkml_transformer.compiler.sql_compiler import SQLCompiler
from linkml_transformer.transformer.transformer import OBJECT_TYPE, Transformer

DICT_OBJ = Dict[str, Any]
DATABASE = Union[str, DuckDBPyConnection]


logger = logging.getLogger(__name__)


@dataclass
class DuckDBTransformer(Transformer):
    """
    A Transformer that works on DuckDB data.
    """

    def map_object(
        self,
        source_obj: OBJECT_TYPE,
        source_type: str = None,
        target_type: str = None,
        **kwargs,
    ) -> OBJECT_TYPE:
        """
        Transform a source object into a target object.

        :param source_obj: source data structure
        :param source_type: source_obj instantiates this (may be class, type, or enum)
        :param target_type: target_obj instantiates this (may be class, type, or enum)
        :return: transformed data, either as type target_type or a dictionary
        """
        # sv = self.source_schemaview
        raise NotImplementedError("DuckDBTransformer.transform")

    def map_database(
        self, source_database: DATABASE, target_database: Optional[DATABASE] = None, **kwargs
    ) -> OBJECT_TYPE:
        """
        Transform source resource.

        :param source_database:
        :param target_database:
        :param kwargs:
        :return:
        """
        import duckdb

        if target_database is None:
            target_database = source_database

        def _connect(db: DATABASE) -> DuckDBPyConnection:
            if isinstance(db, str):
                return duckdb.connect(db)
            return db

        source_connection = _connect(source_database)
        target_connection = _connect(target_database)
        sql_compiler = SQLCompiler()
        source_connection.sql(sql_compiler.create_ddl(self.source_schemaview))
        target_connection.sql(sql_compiler.create_ddl(self.target_schemaview))
        if not self.specification:
            raise ValueError("No specification provided.")
        compiled = sql_compiler.compile(self.specification)
        source_connection.execute(compiled.serialization)



"""
meta-circular interpreter for evaluating python expressions

 - See `<https://stackoverflow.com/questions/2371436/evaluating-a-mathematical-expression-in-a-string>`_

TODO: move to core
"""

import ast
import operator as op

# supported operators
from typing import Any, List, Tuple

operators = {
    ast.Add: op.add,
    ast.Sub: op.sub,
    ast.Mult: op.mul,
    ast.Div: op.truediv,
    ast.Pow: op.pow,
    ast.BitXor: op.xor,
    ast.USub: op.neg,
}
compare_operators = {ast.Eq: op.eq, ast.Lt: op.lt, ast.LtE: op.le, ast.Gt: op.gt, ast.GtE: op.ge}


def eval_conditional(*conds: List[Tuple[bool, Any]]) -> Any:
    """
    >>> x = 10
    >>> eval_conditional((x < 25, 'low'), (x > 25, 'high'), (True, 'low'))
    'low'

    :param subj:
    :return:
    """
    for is_true, val in conds:
        if is_true:
            return val


funcs = {
    "max": (True, max),
    "min": (True, min),
    "len": (True, len),
    "str": (False, str),
    "strlen": (False, len),
    "case": (False, eval_conditional),
}


class UnsetValueError(Exception):
    pass


def eval_expr(expr: str, **kwargs) -> Any:
    """
    Evaluates a given expression, with restricted syntax

    >>> eval_expr('2^6')
    4
    >>> eval_expr('2**6')
    64
    >>> eval_expr('1 + 2*3**(4^5) / (6 + -7)')
    -5.0

    Variables:

    variables can be passed

    >>> eval_expr('{x} + {y}', x=1, y=2)
    3

    Nulls:

    - If a variable is enclosed in {}s then entire expression will eval to None if variable is unset

    >>> print(eval_expr('{x} + {y}', x=None, y=2))
    None


    Functions:

    - only a small set of functions are currently supported. All SPARQL functions will be supported in future

    >>> eval_expr('strlen("a" + "bc")')
    3

    Paths:

    - Expressions such as `person.name` can be used on objects to lookup by attribute/slot
    - Paths can be chained, e.g. `person.address.street`
    - Operations on lists are distributed, e.g `container.persons.name` will return a list of names
    - Similarly `strlen(container.persons.name)` will return a list whose members are the lengths of all names

    :param expr: expression to evaluate
    """
    # if kwargs:
    #    expr = expr.format(**kwargs)
    if expr == "None":
        # TODO: do this as part of parsing
        return None
    try:
        return eval_(ast.parse(expr, mode="eval").body, kwargs)
    except UnsetValueError:
        return None


def eval_(node, bindings=None):
    if isinstance(node, ast.Num):
        return node.n
    elif isinstance(node, ast.Str):
        if "s" in vars(node):
            return node.s
        else:
            return node.value
    elif isinstance(node, ast.Constant):
        return node.value
    elif isinstance(node, ast.NameConstant):
        # can be removed when python 3.7 is no longer supported
        return node.value
    elif isinstance(node, ast.Name):
        if not bindings:
            bindings = {}
        return bindings.get(node.id)
    elif isinstance(node, ast.Subscript):
        if isinstance(node.slice, ast.Index):
            # required for python 3.7
            k = eval_(node.slice.value, bindings)
        else:
            k = eval_(node.slice, bindings)
        v = eval_(node.value, bindings)
        return v[k]
    elif isinstance(node, ast.Attribute):
        # e.g. for person.name, this returns the val of person
        v = eval_(node.value, bindings)

        # lookup attribute, potentially distributing the results over collections
        def _get(obj: Any, k: str, recurse=True) -> Any:
            if isinstance(obj, dict):
                # dicts are treated as collections; distribute results
                if recurse:
                    return [_get(e, k, False) for e in obj.values()]
                else:
                    return obj.get(k)
            elif isinstance(obj, list):
                # attributes are distributed over lists
                return [_get(e, k, False) for e in obj]
            elif obj is None:
                return None
            else:
                return getattr(obj, k)

        return _get(v, node.attr)
    elif isinstance(node, ast.List):
        return [eval_(x, bindings) for x in node.elts]
    elif isinstance(node, ast.Set):
        # sets are not part of the language; we use {x} as notation for x
        if len(node.elts) != 1:
            raise ValueError("The {} must enclose a single variable")
        e = node.elts[0]
        if not isinstance(e, ast.Name):
            raise ValueError("The {} must enclose a variable")
        v = eval_(e, bindings)
        if v is None:
            raise UnsetValueError(f"{e} is not set")
        else:
            return v
    elif isinstance(node, ast.Tuple):
        return tuple([eval_(x, bindings) for x in node.elts])
    elif isinstance(node, ast.Dict):
        return {eval_(k, bindings): eval_(v, bindings) for k, v in zip(node.keys, node.values)}
    elif isinstance(node, ast.Compare):  # <left> <operator> <right>
        if len(node.ops) != 1:
            raise ValueError(f"Must be exactly one op in {node}")
        if type(node.ops[0]) not in compare_operators:
            raise NotImplementedError(f"Not implemented: {node.ops[0]} in {node}")
        py_op = compare_operators[type(node.ops[0])]
        if len(node.comparators) != 1:
            raise ValueError(f"Must be exactly one comparator in {node}")
        right = node.comparators[0]
        return py_op(eval_(node.left, bindings), eval_(right, bindings))
    elif isinstance(node, ast.BinOp):  # <left> <operator> <right>
        return operators[type(node.op)](eval_(node.left, bindings), eval_(node.right, bindings))
    elif isinstance(node, ast.UnaryOp):  # <operator> <operand> e.g., -1
        return operators[type(node.op)](eval_(node.operand, bindings))
    # elif isinstance(node, ast.Match):
    #    # implementing this would restrict python version to 3.10
    #    # https://stackoverflow.com/questions/60208/replacements-for-switch-statement-in-python
    #    raise NotImplementedError(f'Not supported')
    elif isinstance(node, ast.IfExp):
        if eval_(node.test, bindings):
            return eval_(node.body, bindings)
        else:
            return eval_(node.orelse, bindings)
    elif isinstance(node, ast.Call):
        if isinstance(node.func, ast.Name):
            fn = node.func.id
            if fn in funcs:
                takes_list, func = funcs[fn]
                args = [eval_(x, bindings) for x in node.args]
                if isinstance(args[0], list) and not takes_list:
                    return [func(*[x] + args[1:]) for x in args[0]]
                else:
                    return func(*args)
        raise NotImplementedError(f"Call {node.func} not implemented. node = {node}")
    else:
        raise TypeError(node)



from linkml_runtime import SchemaView
from linkml_runtime.linkml_model import SlotDefinition

from linkml_transformer.compiler.compiler import CompiledSpecification, Compiler
from linkml_transformer.datamodel.transformer_model import (
    ClassDerivation,
    SerializationSyntaxType,
    TransformationSpecification,
)

LINKML_TO_SQL_TYPE_MAP = {
    "string": "TEXT",
    "integer": "INTEGER",
    "boolean": "BOOLEAN",
    "float": "REAL",
    "decimal": "REAL",
    "datetime": "TIMESTAMP",
    "date": "DATE",
    "time": "TIME",
    "uri": "TEXT",
    "any": "TEXT",
}


class SQLCompiler(Compiler):
    """
    Compiles a Transformation Specification to SQL CREATE TABLE or VIEW statements.

    Note: this is currently highly geared towards DuckDB.
    """

    add_if_not_exists: bool = True
    new_table_when_transforming: bool = False

    def compile(self, specification: TransformationSpecification) -> CompiledSpecification:
        compiled = CompiledSpecification()
        for cd in specification.class_derivations.values():
            self.compile_class(compiled, cd, specification)
        return compiled

    def compile_class(
        self,
        compiled: CompiledSpecification,
        cd: ClassDerivation,
        specification: TransformationSpecification,
    ) -> None:
        """
        Compile a class derivation to SQL.

        :param compiled: (modified in place)
        :param cd:
        :param specification:
        :return:
        """
        stmt = ""
        if self.new_table_when_transforming:
            stmt += "CREATE TABLE "
            if self.add_if_not_exists:
                stmt += "IF NOT EXISTS "
            stmt += f"{cd.name} \n"
        else:
            stmt += f"INSERT INTO {cd.name} SELECT \n"
        col_trs = []
        for sd in cd.slot_derivations.values():
            col_trs.append(self.compile_slot_derivation(sd))
        if not col_trs:
            return
        stmt += ", \n".join(col_trs)
        if self.new_table_when_transforming:
            stmt += ");"
        stmt += f" FROM {cd.name}"
        compiled.serialization += f"{stmt};\n"

    def compile_slot_derivation(self, sd) -> str:
        expr = sd.populated_from
        if expr is None:
            expr = sd.name
        if sd.stringification:
            syntax = sd.stringification.syntax
            delimiter = sd.stringification.delimiter
            if sd.stringification.reversed:
                pass
            else:
                # duckdb specific?
                if syntax == SerializationSyntaxType.JSON:
                    expr = f"CAST({expr} AS TEXT)"
                elif delimiter:
                    expr = f"STRING_AGG({expr}, '{delimiter}')"
        return f"  {sd.name} AS {expr}"

    def create_ddl(self, schemaview: SchemaView) -> str:
        """
        Create DDL for the entire schema.

        Note core LinkML has a SQL DDL generator, but this is primarily for pure relational models.

        DuckDB allows complex datatypes.

        TODO: move to LinkML core
        :param schemaview:
        :return:
        """
        ddl = []
        for c in schemaview.all_classes().values():
            if c.mixin or c.abstract:
                continue
            col_strs = []
            for s in schemaview.class_induced_slots(c.name):
                col_strs.append(f"  {s.name} {self.sql_type(s, schemaview)}")
            if not col_strs:
                continue
            ddl.append(f"CREATE TABLE IF NOT EXISTS {c.name} (")
            ddl.append(",\n".join(col_strs))
            ddl.append(");")
        return "\n".join(ddl)

    def sql_type(self, slot: SlotDefinition, schemaview: SchemaView) -> str:
        """
        Map LinkML types to DuckDB SQL types.

        :param slot:
        :param schemaview:
        :return:
        """
        typ = "TEXT"
        if slot.range:
            if slot.range in LINKML_TO_SQL_TYPE_MAP:
                typ = LINKML_TO_SQL_TYPE_MAP.get(slot.range, typ)
            elif slot.range in schemaview.all_classes():
                if slot.inlined:
                    typ = "TEXT"
                else:
                    # TODO: consider structs
                    typ = "JSON"

        if slot.multivalued:
            typ = f"{typ}[]"
        return typ



"""Iterate through all examples in a folder testing them for validity.

"""

import glob
import logging
import os
import re
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Mapping, Optional, Union

import yaml
from linkml_runtime import SchemaView
from linkml_runtime.dumpers import yaml_dumper
from pydantic import BaseModel

from linkml_transformer.inference.schema_mapper import SchemaMapper
from linkml_transformer.transformer.object_transformer import ObjectTransformer


class Step(BaseModel):
    source_data: str
    target_data: Optional[str] = None
    source_class: str = None
    target_class: str = None


class Transformation(BaseModel):
    description: Optional[str] = None
    source_schema: str = None
    target_schema: str = None
    transformation_specification: str = None
    steps: List[Step] = None


class Instructions(BaseModel):
    description: Optional[str] = None
    transformations: List[Transformation] = []


@dataclass
class MultiFileTransformer:
    """
    Processes a collection of inputs in one folder.

    Assumes folder structures:

       - {package}
          - source/{source_schema}.yaml  :: the schema to transform from
          - transform/{transform_spec}.transform.yaml :: mapping spec
          - data/{SourceClassName}-{LocalId}.yaml :: data to transform
          - target/{SourceClassName}-{LocalId}.yaml :: expected output data
    """

    source_schema_directory_base: str = field(default_factory=lambda: "source")
    transform_specification_directory_base: str = field(default_factory=lambda: "transform")
    source_data_directory_base: str = field(default_factory=lambda: "data")
    target_schema_directory_base: str = field(default_factory=lambda: "target")
    target_data_directory_base: str = field(default_factory=lambda: "output")

    input_formats: Optional[List[str]] = field(default_factory=lambda: ["yaml"])
    """Expected formats for input data"""

    output_formats: Optional[List[str]] = field(default_factory=lambda: ["yaml"])

    prefix_map: Optional[Mapping[str, str]] = None
    """Custom prefix map, for emitting RDF/turtle."""

    def process_directory(self, root_directory: Union[str, Path], **kwargs):
        """
        Process all transformations in a directory.

        :param root_directory:
        :return:
        """
        instructions = self.infer_instructions(root_directory)
        return self.process_instructions(instructions, root_directory, **kwargs)

    def infer_instructions(self, root_directory: Union[str, Path]) -> Instructions:
        """
        Infer instructions from either explicit yaml or directory layout.
        """
        if isinstance(root_directory, str):
            root_directory = Path(root_directory)
        if not root_directory.exists():
            raise ValueError(f"No such directory {root_directory}")
        instructions_file = root_directory / "instructions.yaml"
        if instructions_file.exists():
            with open(instructions_file) as file:
                return Instructions(**yaml.safe_load(file))
        source_schema_directory = root_directory / self.source_schema_directory_base
        if not source_schema_directory.exists():
            raise ValueError(f"Expected {source_schema_directory}")
        transform_specification_directory = (
            root_directory / self.transform_specification_directory_base
        )
        if not transform_specification_directory.exists():
            raise ValueError(f"Expected {transform_specification_directory}")
        target_schema_directory = root_directory / self.target_schema_directory_base
        input_schemas = glob.glob(os.path.join(str(source_schema_directory), "*.yaml"))
        if not input_schemas:
            raise ValueError(f"Expected schemas in {source_schema_directory}")
        transform_files = glob.glob(
            os.path.join(str(transform_specification_directory), "*.transform.yaml")
        )
        instructions = Instructions(description="auto-inducted")
        for input_schema in input_schemas:
            for transform_file in transform_files:
                tr = Transformation(
                    source_schema=input_schema,
                    transformation_specification=transform_file,
                    steps=[],
                )
                target_schema_base = None
                if len(input_schemas) != 1:
                    # resolve which schema the transform applies to
                    matches = re.match(r"^(\w+)-to-(\w+)\.", transform_file)
                    if not matches:
                        raise ValueError(f"Ambiguous: {transform_file}")
                    src, target_schema_base = matches.group(1, 2)
                    if src not in input_schema:
                        continue
                instructions.transformations.append(tr)
                data_files = glob.glob(
                    os.path.join(str(root_directory / self.source_data_directory_base), "*.yaml")
                )
                for data_file in data_files:
                    if target_schema_base and target_schema_base not in data_file:
                        continue
                    target_data = str(Path(self.target_data_directory_base) / Path(data_file).name)
                    target_data = target_data.replace(".yaml", ".transformed.yaml")
                    step = Step(source_data=data_file, target_data=target_data)
                    step.source_class = Path(data_file).stem.split("-")[-2]
                    tr.steps.append(step)
                target_schemas = glob.glob(os.path.join(str(target_schema_directory), "*.yaml"))
                if len(target_schemas) > 1:
                    target_schemas = [s for s in target_schemas if target_schema_base in s]
                    if len(target_schemas) != 1:
                        raise ValueError(
                            f"Could not determine target schema from: {target_schemas}"
                        )
                if target_schemas:
                    tr.target_schema = target_schemas[0]
                else:
                    tr.target_schema = str(Path(target_schema_directory) / "target.yaml")
                if not tr.steps:
                    raise ValueError(f"Could not infer steps from {data_files}")
                if not tr.transformation_specification:
                    raise ValueError(f"No spec {tr}")
        return instructions

    def process_instructions(
        self,
        instructions: Instructions,
        root_directory: Optional[Union[str, Path]],
        output_directory=None,
        test_mode=False,
    ):
        """
        Process a set of instructions to transform one or more files.

        :param instructions:
        :param root_directory:
        :param output_directory:
        :param test_mode:
        :return:
        """
        if isinstance(root_directory, str):
            root_directory = Path(root_directory)
        if isinstance(output_directory, str):
            root_directory = Path(output_directory)
        if not output_directory:
            output_directory = root_directory
        logging.info(f"Processing: {instructions.description}")
        for tr in instructions.transformations:
            source_schema_path = str(root_directory / tr.source_schema)
            sv = SchemaView(source_schema_path)
            transformer = ObjectTransformer(unrestricted_eval=True)
            transformer.load_source_schema(source_schema_path)
            tr_path = str(root_directory / tr.transformation_specification)
            transformer.load_transformer_specification(tr_path)
            if tr.target_schema:
                target_schema_path = root_directory / tr.target_schema
                mapper = SchemaMapper()
                mapper.source_schemaview = sv
                target_schema_obj = mapper.derive_schema(transformer.specification)
                if not target_schema_path.exists():
                    target_schema_path.parent.mkdir(exist_ok=True, parents=True)
                    yaml_dumper.dump(target_schema_obj, str(target_schema_path))
            for step in tr.steps:
                input_obj = yaml.safe_load(open(str(root_directory / step.source_data)))
                transformer.index(input_obj, step.source_class)
                target_obj = transformer.map_object(input_obj, source_type=step.source_class)
                if step.target_data:
                    out_path = output_directory / step.target_data
                    out_path.parent.mkdir(parents=True, exist_ok=True)
                    if out_path.exists():
                        line1 = open(str(out_path)).readline()
                        overwrite = "OVERWRITE" in line1
                        with open(str(out_path)) as f:
                            compare_obj = yaml.safe_load(f)
                            if compare_obj != target_obj:
                                if test_mode and not overwrite:
                                    raise ValueError(
                                        f"Output different than expected: {compare_obj}"
                                    )
                                else:
                                    logging.warning(f"Different: {compare_obj}")
                    with open(str(out_path), "w", encoding="utf-8") as f:
                        yaml_str = yaml.dump(target_obj, sort_keys=False)
                        f.write(yaml_str)



from dataclasses import dataclass

from jinja2 import Environment, FileSystemLoader

from linkml_transformer.compiler.compiler import CompiledSpecification, Compiler
from linkml_transformer.compiler.templates import TEMPLATE_DIR
from linkml_transformer.datamodel.transformer_model import TransformationSpecification


@dataclass
class J2BasedCompiler(Compiler):
    """
    Compiles a Transformation Specification using a Jinja2 template.
    """

    template_dir: str = None
    """The directory containing the Jinja2 template."""

    template_name: str = None
    """The name of the Jinja2 template."""

    def compile(self, specification: TransformationSpecification) -> CompiledSpecification:
        template_dir = self.template_dir
        if not template_dir:
            template_dir = TEMPLATE_DIR
        if not template_dir:
            raise ValueError("template_dir must be set")
        loader = FileSystemLoader(template_dir)
        env = Environment(loader=loader, autoescape=True)
        if not self.template_name:
            raise ValueError("template_name must be set")
        template = env.get_template(self.template_name)
        rendered = template.render(
            spec=specification,
        )
        return CompiledSpecification(serialization=rendered)



import re
from dataclasses import dataclass
from typing import List, Optional, Tuple

from graphviz import Digraph
from linkml_runtime import SchemaView
from pydantic import BaseModel

from linkml_transformer.compiler.compiler import CompiledSpecification, Compiler
from linkml_transformer.datamodel.transformer_model import TransformationSpecification


class Record(BaseModel):
    """
    A simplified representation of a class, UML-style.
    """

    name: str
    source: str
    fields: List[Tuple[str, str]] = []

    @property
    def id(self):
        return f"{self.source}{self.name}"

    def __str__(self):
        return (
            f"""<
        <TABLE BORDER="0" CELLBORDER="1" CELLSPACING="0">
            <TR><TD><B>{self.name}</B></TD></TR>
        """
            + "".join(
                [
                    f'<TR><TD SIDES="LRB" PORT="{f[0]}">{f[0]} : {f[1]}</TD></TR>'
                    for f in self.fields
                ]
            )
            + """
        </TABLE>>"""
        )


@dataclass
class GraphvizObject(CompiledSpecification):
    digraph: Digraph = None

    def render(self, file_path: str, format="png", view=False) -> None:
        """
        Renders a graphviz graph to a file.
        :param file_path:
        :return:
        """
        self.digraph.render(file_path, format=format, view=view)


class GraphvizCompiler(Compiler):
    """
    Compiles a Transformation Specification to GraphViz.
    """

    def compile(
        self, specification: TransformationSpecification, elements: Optional[List[str]] = None
    ) -> GraphvizObject:
        dg = Digraph(comment="UML Class Diagram", format="png")
        dg.attr(rankdir="LR")  # Set graph direction from left to right
        target_schemaview = self.derived_target_schemaview(specification)
        source_schemaview = self.source_schemaview

        records = []
        records += self.add_records(source_schemaview, "source")
        records += self.add_records(target_schemaview, "target")

        for record in records:
            dg.node(record.id, str(record), shape="plaintext")

        # Define the class nodes with fields in UML format using HTML-like labels
        # for precise control over the stacking of the fields
        for target_cn, cd in specification.class_derivations.items():
            source_cn = cd.populated_from
            if source_cn is None:
                source_cn = cd.name
            if elements is not None and target_cn not in elements:
                continue
            source_record = Record(name=source_cn, source="source")
            target_record = Record(name=target_cn, source="target")
            for sd in cd.slot_derivations.values():
                target_slot = sd.name
                target_id = f"{target_record.id}:{target_slot}"
                source_slot = sd.populated_from
                if source_slot:
                    source_id = f"{source_record.id}:{source_slot}"
                    dg.edge(source_id, target_id)
                elif sd.expr:
                    # TODO: do this in a less hacky way
                    tokens = re.findall(r"\w+", sd.expr)
                    for token in tokens:
                        if token not in source_schemaview.all_slots():
                            continue
                        dg.edge(f"{source_record.id}:{token}", target_id, style="dashed")
        return GraphvizObject(digraph=dg, serialization=dg.source)

    def add_records(self, schemaview: SchemaView, source: str) -> List[Record]:
        records = []
        for cn in schemaview.all_classes():
            record = Record(name=cn, source=source)
            for induced_slot in schemaview.class_induced_slots(cn):
                record.fields.append((induced_slot.name, induced_slot.range))
            records.append(record)
        return records



import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Optional, Type, Union

import yaml
from linkml_runtime import SchemaView
from linkml_runtime.dumpers import yaml_dumper
from linkml_runtime.linkml_model import SchemaDefinition
from linkml_runtime.processing.referencevalidator import ReferenceValidator
from linkml_runtime.utils.introspection import package_schemaview

from linkml_transformer import ObjectTransformer
from linkml_transformer.datamodel.transformer_model import TransformationSpecification
from linkml_transformer.inference.inverter import TransformationSpecificationInverter
from linkml_transformer.inference.schema_mapper import SchemaMapper
from linkml_transformer.transformer.transformer import Transformer

logger = logging.getLogger(__name__)


@dataclass
class Session:
    """
    A wrapper object for a transformer session.

    TODO:

    - rename to Manager?
    - consolidate configuration
    - include source and target database

        - current spec, src_sv, tgt_sv all live in both this class and transformer
    """

    transformer_specification: Optional[TransformationSpecification] = None
    source_schemaview: Optional[SchemaView] = None
    transformer: Optional[Transformer] = None
    object_transformer: Optional[ObjectTransformer] = None
    schema_mapper: Optional[SchemaMapper] = None
    _target_schema: Optional[SchemaDefinition] = None
    _target_schemaview: Optional[SchemaView] = None

    def set_transformer_specification(
        self, specification: Optional[Union[TransformationSpecification, dict, str, Path]] = None
    ):
        if isinstance(specification, Path):
            specification = str(specification)
        if isinstance(specification, TransformationSpecification):
            self.transformer_specification = specification
        elif isinstance(specification, dict):
            # TODO: centralize this code
            normalizer = ReferenceValidator(
                package_schemaview("linkml_transformer.datamodel.transformer_model")
            )
            normalizer.expand_all = True
            specification = normalizer.normalize(specification)
            self.transformer_specification = TransformationSpecification(**specification)
        elif isinstance(specification, str):
            if "\n" in specification:
                obj = yaml.safe_load(specification)
            else:
                obj = yaml.safe_load(open(specification))
            self.set_transformer_specification(obj)

    def set_source_schema(self, schema: Union[str, Path, dict, SchemaView, SchemaDefinition]):
        """
        Sets the schema from a path or SchemaView object.
        """
        if isinstance(schema, str):
            sv = SchemaView(schema)
        elif isinstance(schema, Path):
            sv = SchemaView(str(schema))
        elif isinstance(schema, dict):
            sv = SchemaView(yaml_dumper.dumps(schema))
        elif isinstance(schema, SchemaView):
            sv = schema
        elif isinstance(schema, SchemaDefinition):
            sv = SchemaView(schema)
        else:
            raise ValueError(f"Unsupported schema type: {type(schema)}")
        self.source_schemaview = sv
        self._target_schema = None

    def set_transformer(
        self,
        transformer: Optional[Union[Transformer, Type[Transformer]]],
        **kwargs,
    ):
        if isinstance(transformer, Type):
            transformer = transformer()
        transformer.specification = self.transformer_specification
        self.transformer = transformer

    def set_object_transformer(
        self,
        transformer: Optional[
            Union[ObjectTransformer, TransformationSpecification, dict, str, Path]
        ] = None,
    ):
        if transformer is None:
            if self.object_transformer is not None:
                logger.info("No change")
                return
            else:
                logger.warning("No transformer specified")
                return
        if transformer is not None:
            if isinstance(transformer, ObjectTransformer):
                self.object_transformer = transformer
            else:
                self.set_transformer_specification(transformer)
                self.object_transformer = ObjectTransformer()
                self.object_transformer.specification = self.transformer_specification
        self._target_schema = None

    @property
    def target_schema(self) -> SchemaDefinition:
        if self._target_schema is None:
            if not self.schema_mapper:
                self.schema_mapper = SchemaMapper(source_schemaview=self.source_schemaview)
            self._target_schema = self.schema_mapper.derive_schema(self.transformer_specification)
        return self._target_schema

    @property
    def target_schemaview(self) -> SchemaView:
        if self._target_schemaview is None:
            # TODO: simplify
            self._target_schemaview = SchemaView(yaml_dumper.dumps(self.target_schema))
        return self._target_schemaview

    def transform(self, obj: dict, **kwargs) -> dict:
        if self.object_transformer is None:
            raise ValueError("No transformer specified")
        if not self.object_transformer.source_schemaview:
            self.object_transformer.source_schemaview = self.source_schemaview
        return self.object_transformer.map_object(obj, **kwargs)

    def reverse_transform(self, obj: dict, **kwargs) -> dict:
        inv_spec = self.invert()
        reverse_transformer = ObjectTransformer()
        reverse_transformer.specification = inv_spec
        reverse_transformer.source_schemaview = SchemaView(yaml_dumper.dumps(self.target_schema))
        return reverse_transformer.map_object(obj, **kwargs)

    def invert(self, in_place=False) -> TransformationSpecification:
        """
        Invert the transformer specification.
        """
        inverter = TransformationSpecificationInverter(
            source_schemaview=self.source_schemaview,
            target_schemaview=SchemaView(yaml_dumper.dumps(self.target_schema)),
        )
        inv_spec = inverter.invert(self.transformer_specification)
        if in_place:
            raise NotImplementedError
        return inv_spec

    def graphviz(self, **kwargs) -> Any:
        """
        Return a graphviz representation of the schema.
        """
        from linkml_transformer.compiler.graphviz_compiler import GraphvizCompiler

        gc = GraphvizCompiler(source_schemaview=self.source_schemaview)
        compiled = gc.compile(self.transformer_specification)
        return compiled.digraph



from linkml_transformer.compiler.compiler import CompiledSpecification, Compiler
from linkml_transformer.datamodel.transformer_model import TransformationSpecification


class R2RMLCompiler(Compiler):
    """
    Compiles a Transformation Specification to R2RML.

    """

    def compile(self, specification: TransformationSpecification) -> CompiledSpecification:
        raise NotImplementedError



import logging
import re
from copy import copy
from dataclasses import dataclass, field
from typing import Optional

from linkml_runtime import SchemaView

from linkml_transformer.datamodel.transformer_model import (
    ClassDerivation,
    CollectionType,
    EnumDerivation,
    PermissibleValueDerivation,
    SlotDerivation,
    TransformationSpecification,
    UnitConversionConfiguration,
)

logger = logging.getLogger(__name__)


class NonInvertibleSpecification(ValueError):
    pass


@dataclass
class TransformationSpecificationInverter:
    """
    Invert a transformation specification.
    """

    source_schemaview: SchemaView = None
    """The source schema for the forward transformation.
    Note this becomes the target schema for the generated inverse transformation,
    because the goal of the inverted transformation is to map back to the original source."""

    target_schemaview: SchemaView = None
    """The target schema for the forward transformation.
    Note this becomes the source schema for the generated inverse transformation."""

    strict: bool = field(default=True)

    def invert(self, spec: TransformationSpecification):
        """
        Invert a transformation specification.

        :param spec:
        :return:
        """
        logger.info("Inverting specification")
        inverted_spec = TransformationSpecification()
        for cd in spec.class_derivations.values():
            inverted_cd = self.invert_class_derivation(cd, spec)
            inverted_spec.class_derivations[inverted_cd.name] = inverted_cd
        for ed in spec.enum_derivations.values():
            inverted_ed = self.invert_enum_derivation(ed, spec)
            inverted_spec.enum_derivations[inverted_ed.name] = inverted_ed
        return inverted_spec

    def invert_class_derivation(self, cd: ClassDerivation, spec: TransformationSpecification):
        """
        Invert a class derivation.

        :param cd:
        :param spec:
        :return:
        """
        inverted_cd = ClassDerivation(
            name=cd.populated_from if cd.populated_from else cd.name, populated_from=cd.name
        )
        for sd in cd.slot_derivations.values():
            inverted_sd = self.invert_slot_derivation(sd, cd, spec)
            if inverted_sd:
                inverted_cd.slot_derivations[inverted_sd.name] = inverted_sd
            else:
                if self.strict:
                    raise NonInvertibleSpecification(f"Cannot invert slot derivation: {sd.name}")
        return inverted_cd

    def invert_enum_derivation(self, ed: EnumDerivation, spec: TransformationSpecification):
        """
        Invert an enum derivation.

        :param ed:
        :param spec:
        :return:
        """
        inverted_ed = EnumDerivation(
            name=ed.populated_from if ed.populated_from else ed.name, populated_from=ed.name
        )
        if inverted_ed.expr:
            raise NonInvertibleSpecification("TODO: invert enum derivation with expression")
        for pv_deriv in ed.permissible_value_derivations.values():
            inverted_pv_deriv = PermissibleValueDerivation(
                name=pv_deriv.populated_from if pv_deriv.populated_from else pv_deriv.name,
                populated_from=pv_deriv.name,
            )
            inverted_ed.permissible_value_derivations[inverted_pv_deriv.name] = inverted_pv_deriv
        return inverted_ed

    def invert_slot_derivation(
        self, sd: SlotDerivation, cd: ClassDerivation, spec: TransformationSpecification
    ) -> Optional[SlotDerivation]:
        """
        Invert a slot derivation.

        :param sd:
        :param cd:
        :param spec:
        :return:
        """
        populated_from = sd.populated_from
        if sd.expr:
            if re.match(r"^\w+$", sd.expr):
                populated_from = sd.expr
            else:
                if not self.strict:
                    return None
                # TODO: add logic for reversible expressions
                raise NonInvertibleSpecification(
                    f"Cannot invert expression {sd.expr} in slot derivation: {sd.name}"
                )
        if not populated_from:
            # use defaults. TODO: decide on semantics of defaults
            populated_from = sd.name
            # raise NonInvertibleSpecification(f"No populate_from or expr in slot derivation: {sd.name}")
        inverted_sd = SlotDerivation(name=populated_from, populated_from=sd.name)
        # source_cls_name = spec.class_derivations[cd.populated_from].name
        source_cls_name = cd.populated_from
        if (
            source_cls_name is None or source_cls_name in self.source_schemaview.all_classes()
        ) and sd.populated_from:
            source_slot = self.source_schemaview.induced_slot(sd.populated_from, source_cls_name)
        else:
            source_slot = None
        if sd.range:
            # source_slot = self.source_schemaview.induced_slot(sd.populated_from, source_cls_name)
            inverted_sd.range = source_slot.range
            if source_slot.range in self.source_schemaview.all_classes():
                id_slot = self.source_schemaview.get_identifier_slot(
                    source_slot.range, use_key=True
                )
                if id_slot:
                    inverted_sd.dictionary_key = id_slot.name
        if source_slot and source_slot.multivalued:
            if source_slot.inlined_as_list:
                inverted_sd.cast_collection_as = CollectionType.MultiValuedList
            elif source_slot.inlined:
                if source_slot.range in self.source_schemaview.all_classes():
                    id_slot = self.source_schemaview.get_identifier_slot(
                        source_slot.range, use_key=True
                    )
                    if id_slot:
                        inverted_sd.cast_collection_as = CollectionType.MultiValuedDict
                        inverted_sd.dictionary_key = id_slot.name
        if sd.unit_conversion:
            source_slot = self.source_schemaview.induced_slot(sd.populated_from, source_cls_name)
            target_unit = None
            target_unit_scheme = None
            if source_slot.unit is not None:
                for p in ["ucum_code", "symbol"]:
                    target_unit = getattr(source_slot.unit, p, None)
                    if target_unit is not None:
                        target_unit_scheme = p
                        break
            inverted_uc = UnitConversionConfiguration(
                target_unit=target_unit, target_unit_scheme=target_unit_scheme
            )
            # if sd.unit_conversion.target_unit:
            #    inverted_uc.source_unit = sd.unit_conversion.target_unit
            if sd.unit_conversion.source_unit_slot:
                inverted_uc.target_unit_slot = sd.unit_conversion.source_unit_slot
            if sd.unit_conversion.source_magnitude_slot:
                inverted_uc.target_magnitude_slot = sd.unit_conversion.source_magnitude_slot
            if sd.unit_conversion.target_unit:
                inverted_uc.source_unit = sd.unit_conversion.target_unit
            inverted_sd.unit_conversion = inverted_uc
        if sd.stringification:
            inverted_sd.stringification = copy(sd.stringification)
            inverted_sd.stringification.reversed = not inverted_sd.stringification.reversed
        return inverted_sd



from dataclasses import dataclass

from linkml_transformer.compiler.j2_based_compiler import J2BasedCompiler


@dataclass
class MarkdownCompiler(J2BasedCompiler):
    """
    Compiles a Transformation Specification to Markdown.
    """

    template_name: str = "markdown.j2"



"""
Compilers are responsible for compiling a transformation specification into an alternative representation.

This is the opposite of an importer.

For example:

- LinkML-Transformer Specifications to R2RML
- LinkML-Transformer Specifications to awk scripts
- LinkML-Transformer Specifications to SQL
- LinkML-Transformer Specifications to Python (OO)
- LinkML-Transformer Specifications to Pandas
- LinkML-Transformer Specifications to Hamilton
"""

from abc import ABC
from dataclasses import dataclass, field
from types import ModuleType
from typing import Iterator, Optional

from linkml_runtime import SchemaView
from linkml_runtime.dumpers import yaml_dumper
from linkml_runtime.utils.compile_python import compile_python

from linkml_transformer.datamodel.transformer_model import TransformationSpecification
from linkml_transformer.inference.schema_mapper import SchemaMapper


@dataclass
class CompiledSpecification:
    serialization: str = field(default="")

    _module: Optional[ModuleType] = None

    @property
    def module(self) -> ModuleType:
        if not self._module:
            self._module = compile_python(self.serialization)
        return self._module


@dataclass
class Compiler(ABC):
    """
    Base class for all compilers.

    A compiler will compile a transformation specification into
    an alternative representation.

    An example compiler would be a R2RML compiler.

    Note: Compilers and Importers will in general be implemented by providing
    mapping specifications
    """

    source_schemaview: SchemaView = None
    """A view over the schema describing the source."""

    source_python_module: str = None
    """The python module containing the source classes."""

    target_python_module: str = None
    """The python module containing the target classes."""

    def compile(self, specification: TransformationSpecification) -> CompiledSpecification:
        """
        Transform source object into an instance of the target class.

        :param specification:
        :return:
        """
        s = self._compile_header(specification)
        for chunk in self._compile_iterator(specification):
            s += chunk
        return CompiledSpecification(serialization=s)

    def _compile_header(self, specification: TransformationSpecification) -> str:
        return ""

    def _compile_iterator(self, specification: TransformationSpecification) -> Iterator[str]:
        raise NotImplementedError

    def derived_target_schemaview(self, specification: TransformationSpecification):
        """
        Returns a view over the target schema, including any derived classes.
        """
        mapper = SchemaMapper(source_schemaview=self.source_schemaview)
        return SchemaView(yaml_dumper.dumps(mapper.derive_schema(specification)))



from typing import Any

from linkml_runtime import SchemaView


class DynObj:
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __str__(self) -> str:
        return str(vars(self))

    def __getattr__(self, p: str) -> Any:
        return vars(self).get(p, None)


def dynamic_object(obj: dict, sv: SchemaView, target: str):
    """
    Generate a dynamic object from a dict.

    :param obj:
    :param sv:
    :param target:
    :return:
    """
    if target in sv.all_enums():
        return obj
    if target in sv.all_types():
        return obj
    if not isinstance(obj, dict):
        return obj
    attrs = {}
    for k, v in obj.items():
        slot = sv.induced_slot(k, target)
        rng = slot.range
        if slot.multivalued:
            if isinstance(v, list):
                v = [dynamic_object(x, sv, rng) for x in v]
            if isinstance(v, dict):
                v = {xk: dynamic_object(x, sv, rng) for xk, x in v.items()}
                id_slot = sv.get_identifier_slot(slot.range)
                for k1, v1 in v.items():
                    setattr(v1, id_slot.name, k1)
        else:
            v = dynamic_object(v, sv, rng)
        attrs[k] = v
    cls = type(target, (DynObj,), {})
    return cls(**attrs)



from copy import deepcopy
from dataclasses import dataclass
from typing import Iterator

from jinja2 import Template

from linkml_transformer.compiler.compiler import Compiler
from linkml_transformer.datamodel.transformer_model import (
    ClassDerivation,
    TransformationSpecification,
)
from linkml_transformer.inference.inference import induce_missing_values

CD_TEMPLATE = """
{% macro gen_slot_derivation_value(sd, var) -%}
{%- if sd.range -%}
derive_{{ sd.range }}({{ var }})
{%- else -%}
{%- if var is not none -%}
{{ var }}
{%- else -%}
None
{%- endif -%}
{%- endif -%}
{%- endmacro %}
{% macro gen_slot_derivation(sd, force_singlevalued=False) -%}
{%- if not force_singlevalued and sd.populated_from and induced_slots[sd.populated_from].multivalued -%}
 [ {{ gen_slot_derivation_value(sd, "x") }} for x in {{ gen_slot_derivation(sd, force_singlevalued=True) }} ]
{%- else -%}
 {%- if sd.populated_from -%}
  source_object.{{ sd.populated_from }}
 {%- elif sd.expr -%}
  {%- if '\n' in sd.expr -%}
 gen_{{ sd.name }}(source_object)
  {%- elif '{' in sd.expr and '}' in sd.expr -%}
  {{ sd.expr|replace('{', '')|replace('}', '') }}
  {%- else -%}
  {{ sd.expr }}
  {%- endif -%}
 {%- else -%}
  None
 {%- endif -%}
{%- endif -%}
{%- endmacro %}
{% macro gen_slot_derivation_defs(sd) -%}
{% if sd.expr and '\n' in sd.expr %}

    def gen_{{ sd.name }}(src):
        target = None
   {%- for line in sd.expr.split('\n') %}
        {{ line }}
   {%- endfor -%}
        return target
{% endif %}
{%- endmacro %}
def derive_{{ cd.name }}(
        source_object: {{ source_module }}.{{ cd.populated_from }}
    ) -> {{ target_module }}.{{ cd.name }}:
    # assign slots
{%- for slot in source_slots %}
    {{ slot.name }} = source_object.{{ slot.name }}
{%- endfor %}
{%-  for sd in cd.slot_derivations.values() -%}
    {{  gen_slot_derivation_defs(sd) }}
{%-  endfor %}

    return tgt.{{ cd.name }}(
        {%- for sd in cd.slot_derivations.values() %}
        {{ sd.name }}={{ gen_slot_derivation(sd) }},
        {%- endfor %}
    )
"""


@dataclass
class PythonCompiler(Compiler):
    """
    Compiles a Transformation Specification to Python code.
    """

    def _compile_header(self, specification: TransformationSpecification) -> str:
        s = ""
        if self.source_python_module:
            s += f"import {self.source_python_module} as src\n"
        if self.target_python_module:
            s += f"import {self.target_python_module} as tgt\n"
        s += "\nNULL = None\n\n"
        return s

    def _compile_iterator(self, specification: TransformationSpecification) -> Iterator[str]:
        specification = deepcopy(specification)
        induce_missing_values(specification, self.source_schemaview)
        for cd in specification.class_derivations.values():
            yield from self._compiled_class_derivations_iter(cd)

    def _compiled_class_derivations_iter(self, cd: ClassDerivation) -> Iterator[str]:
        sv = self.source_schemaview
        if cd.populated_from:
            populated_from = cd.populated_from
        else:
            populated_from = cd.name
        if populated_from not in sv.all_classes():
            return
        induced_slots = {s.name: s for s in sv.class_induced_slots(populated_from)}
        t = Template(CD_TEMPLATE)
        yield t.render(
            cd=cd,
            source_module="src",
            target_module="tgt",
            induced_slots=induced_slots,
            schemaview=sv,
            source_slots=sv.class_induced_slots(populated_from),
        )
